#pragma once

#include <Windows.h>
#include <d3d11.h>
#include <string>
#include <atomic>

#include "Buffer.h"


enum class CaptureMode;


class Window
{
friend class WindowManager;
public:
    struct Data1
    {
        BOOL isDesktop;
        HWND hWnd;
        HMONITOR hMonitor;
        HWND hOwner;
        RECT windowRect;
        RECT clientRect;
        UINT zOrder;
    };

    struct Data2
    {
        HWND hParent;
        HINSTANCE hInstance;
        DWORD processId;
        DWORD threadId;
        std::wstring title;
        std::string className;
        BOOL isAltTabWindow;
        BOOL isApplicationFrameWindow;
        BOOL isUWP;
        BOOL isBackground;
    };

    Window(int id, const Data1 &data);
    ~Window();

    void SetData(const Data1& data);

    int GetId() const;
    int GetParentId() const;
    HWND GetWindowHandle() const;
    HMONITOR GetMonitorHandle() const;
    HWND GetOwnerHandle() const;
    HWND GetParentHandle() const;
    HINSTANCE GetInstance() const;
    DWORD GetProcessId() const;
    DWORD GetThreadId() const;

    const RECT & GetWindowRect() const;
    const RECT & GetClientRect() const;

    UINT GetX() const;
    UINT GetY() const;
    UINT GetWidth() const;
    UINT GetHeight() const;
    UINT GetClientWidth() const;
    UINT GetClientHeight() const;
    UINT GetZOrder() const;
    BYTE* GetBuffer() const;
    UINT GetTextureWidth() const;
    UINT GetTextureHeight() const;
    UINT GetTextureOffsetX() const;
    UINT GetTextureOffsetY() const;
    UINT GetIconWidth() const;
    UINT GetIconHeight() const;

    const std::wstring& GetTitle() const;
    const std::string& GetClass() const;

    void SetWindowTexture(ID3D11Texture2D* ptr);
    ID3D11Texture2D* GetWindowTexture() const;

    void SetIconTexture(ID3D11Texture2D* ptr);
    ID3D11Texture2D* GetIconTexture() const;

    void SetCaptureMode(CaptureMode mode);
    CaptureMode GetCaptureMode() const;

    void SetCursorDraw(bool draw);
    bool GetCursorDraw() const;

    UINT GetPixel(int x, int y) const;
    bool GetPixels(BYTE* output, int x, int y, int width, int height) const;

    void RequestUpdateTitle();

    void Capture();
    void Upload();
    void Render();

    void CaptureIcon();
    void UploadIcon();
    void RenderIcon();

    bool IsAltTab() const;
    bool IsDesktop() const;
    BOOL IsWindow() const;
    BOOL IsVisible() const;
    BOOL IsEnabled() const;
    BOOL IsUnicode() const;
    BOOL IsZoomed() const;
    BOOL IsIconic() const;
    BOOL IsHungUp() const;
    BOOL IsTouchable() const;
    BOOL IsApplicationFrameWindow() const;
    BOOL IsUWP() const;
    BOOL IsBackground() const;

    bool IsWindowsGraphicsCaptureAvailable() const;

private:
    void InitTexture();
    void UpdateFrameCount();
    void UpdateTitle();
    void UpdateIsBackground();
    bool IsJustAdded() const;

    const int id_ = -1;
    int parentId_ = -1;
    Data1 data1_ = {};
    Data2 data2_ = {};
    std::shared_ptr<class WindowTexture> windowTexture_;
    std::shared_ptr<class IconTexture> iconTexture_;

    int frameCount_ = 0;

    std::atomic<bool> hasTitleUpdateRequested_ = false;
    std::atomic<bool> hasNewWindowTextureCaptured_ = false;
    std::atomic<bool> hasNewWindowTextureUploaded_ = false;
    std::atomic<bool> hasNewIconTextureUploaded_ = false;
    std::atomic<bool> isAlive_ = true;
};
